"""credits.py - a context for displaying a game's credits

Slightly fancier example with some spritefx.
"""

import random

import pygame

import clock
import context
import settings
import spritefx
import tweening


class FloatingText(spritefx.TextSprite):
    def update(self, dt):
        self.x += self.dx
        self.dy += 9.8337 * dt
        self.y += self.dy


class Credits(context.Context):
    def __init__(self):
        context.Context.__init__(self, clock.FixedDriver(self.update, settings.time_step))
        self.clock.new_schedule(self.draw, settings.draw_step)

        self.bgcolor = 0, 0, 0

        self.sprites = []

    def enter(self):
        # see context.Context
        def driftx_end(*args):
            args[0].end_driftx()
            self.clock.new_schedule(self.new_floating_text, 1.0 / 8.0)

        x, y = settings.screen_rect.center
        s = spritefx.TextSprite('Press Escape', (x, y), theme_name='game', anchor='center')
        s.scale_effect(start_scale=0.1, end_scale=1.0, duration=2.5, tween_function=tweening.ease_out_quad)
        s.driftx_effect(start_x=x - 350, end_x=x, duration=3.5, tween_function=tweening.ease_out_quad, cb_end=driftx_end)
        s.drifty_effect(start_y=y - 300, end_y=y - 50, duration=2.5, tween_function=tweening.ease_out_bounce)
        self.sprites.append(s)

    def exit(self):
        # see context.Context
        spritefx.clear()

    def suspend(self):
        # see context.Context
        pass

    def resume(self):
        # see context.Context
        pass

    def quit_context(self):
        self.pop()

    def new_floating_text(self, *args):
        # above or below center text
        signx = random.choice((1, -1))
        signy = random.choice((1, -1))
        # starting pos, text selection, and create the sprite
        x, y = settings.screen_rect.center
        y -= {-1: 100, 1: 0}[signy]
        text = random.choice('Score Yay Win Party Cheer Credits'.split())
        if random.choice((True, False)):
            text = text.upper()
        s = FloatingText(text, (x, y), expire=2.0, theme_name='floatingtext', anchor='center')
        # apply fade effect
        s.fade_effect(start_alpha=255, end_alpha=32, duration=1.0, tween_function=tweening.ease_in_quad)
        # random speed: shooting upward if above center text, falling downward if below
        s.dx = random.randrange(120) * signx * settings.time_step
        g = 9.8337 * settings.time_step
        if signy == -1:
            # above, shooting upward
            s.dy = -random.randrange(24, 38) * g
        else:
            # below, falling
            s.dy = random.randrange(-16, 0) * random.choice((0, 0, 0, 1)) * g

    def update(self, dt):
        self.get_events()
        spritefx.update_tweens(dt)
        for s in spritefx.texts:
            s.update(dt)
    
    def draw(self, sched):
        screen = settings.screen
        screen.fill(self.bgcolor)
        for s in self.sprites:
            s.draw(screen)
        for s in spritefx.texts:
            s.draw(screen)
        pygame.display.flip()

    def on_key_down(self, unicode, key, mod):
        # see context.Context.get_events()
        if key == pygame.K_ESCAPE:
            self.quit_context()

    def on_quit(self):
        # see context.Context.get_events()
        self.quit_context()
