"""game.py - a context for running a game session or level
"""

import gc
from itertools import chain
import random
import timeit

import pygame

import clock
import context
import settings

import binspace
import binspacealternatives
import spatialhash
import hudlight
import hudthemes
from geometry import point_on_circumference,angle_of
from vec2d import Vec2d


class Sprite(object):
    lives = 1
    image = None

    def __init__(self, rect):
        self.reset_image = self.image
        self.rect = rect
        self.x = rect.x
        self.y = rect.y
        self.dirx = 0
        self.diry = 0
        self.speed = 1
        self.dim = [self, rect.x, rect.y, rect.w // 2]   # <<< faster than a method call

    def update(self, dt):
        pass


class Player(Sprite):
    number = 0
    player_box = None
    corners = 'topleft', 'topright', 'bottomleft', 'bottomright'

    def update(self, dt):
        corner = self.corners[self.number]
        setattr(self.rect, corner, getattr(self.player_box, corner))
        # self.dim[1:3] = self.rect.topleft   # <<< faster than creating a new tuple 1200 times per game loop


class Bullet(Sprite):
    def update(self, dt):
        dirx = self.dirx
        diry = self.diry
        self.image = self.reset_image
        rect = self.rect
        newx = self.x + dirx
        newy = self.y + diry
        world_rect = settings.world_rect
        hit_edge = False
        if newx < 0 or newx + self.rect.w >= world_rect.right:
            self.dirx = -self.dirx
            hit_edge = True
        if newy < 0 or newy + self.rect.h >= world_rect.bottom:
            self.diry = -self.diry
            hit_edge = True
        rect.x = self.x = newx
        rect.y = self.y = newy
        # self.dim[1:3] = rect.topleft   # <<< faster than creating a new tuple 1200 times per game loop

        if hit_edge:
            self.lives -= 1
            # todo: play a sound


class BaddyBullet(Bullet):
    pass


class BigBaddyBullet(BaddyBullet):
    images = []


class FriendlyBullet(Bullet):
    pass


class Boom(Sprite):
    images = []
    elapsed = 0.0000
    i = 0
    duration = 1.33

    def __init__(self, rect):
        super(Boom, self).__init__(rect)
        self.x, self.y = rect.center
        self.speed = 7

    def update(self, dt):
        self.elapsed += dt
        if self.elapsed > self.duration:
            self.lives = 0
        else:
            self.x += self.dirx * self.speed * dt
            self.y += self.diry * self.speed * dt
            self.speed *= 0.9
            i = int(len(self.images) * self.elapsed / self.duration)
            if i != self.i:
                self.image = self.images[i]
                self.rect.size = self.image.get_size()
            self.rect.center = self.x, self.y

        # self.dim[1:3] = self.rect.topleft   # <<< faster than creating a new tuple 1200 times per game loop


class Game(context.Context):
    def __init__(self):
        context.Context.__init__(self, clock.FixedDriver(self.update, settings.time_step))
        self.sched_draw = self.clock.new_schedule(self.draw, settings.draw_step, keep_history=True)

        self.bgcolor = pygame.Color(0, 0, 0)

        self.players = []
        self.baddy_bullets = []
        self.friendly_bullets = []
        # when sprite.elapsed > 0.55 it is moved from booms to cold_booms to limit chained explosions
        self.booms = []
        self.cold_booms = []

        self.dead_baddy = set()
        self.dead_baddy_boom = set()
        self.dead_friendly = set()
        self.dead_player = set()

        Player.image = pygame.Surface((20, 20))
        Player.image.fill(pygame.Color('mediumpurple3'))
        BaddyBullet.image = pygame.Surface((5, 5))
        BaddyBullet.image.fill(pygame.Color('orangered2'))
        for size in ((10, 19), (15, 35), (19, 31), (23, 29)):
            for i in range(2):
                img = pygame.Surface(size)
                BigBaddyBullet.images.append(img)
                img.set_alpha(128)
                pygame.draw.rect(img, pygame.Color('orangered4'), img.get_rect())
                pygame.draw.rect(img, pygame.Color('orange'), img.get_rect(), 1)
                size = size[1], size[0]
        FriendlyBullet.image = pygame.Surface((5, 5))
        FriendlyBullet.image.fill(pygame.Color('darkviolet'))

        def make_boom(r, alpha):
            w = r * 2
            image = pygame.Surface((w, w))
            pygame.draw.circle(image, pygame.Color('yellow3' if r > 1 else 'yellow'), (r, r), r)
            if r > 3:
                pygame.draw.circle(image, pygame.Color('orangered'), (r, r), r, 2)
                pygame.draw.circle(image, pygame.Color('orange3'), (r, r), r, 1)
            image.set_colorkey((0, 0, 0))
            image.set_alpha(alpha)
            return image
        Boom.images.append(make_boom(1, int(255)))
        Boom.images.append(Boom.images[-1])
        Boom.images.append(Boom.images[-1])
        Boom.images.append(Boom.images[-1])
        Boom.images.append(Boom.images[-1])
        Boom.images.append(Boom.images[-1])
        Boom.images.append(Boom.images[-1])
        Boom.images.append(Boom.images[-1])
        Boom.images.append(make_boom(2, int(255)))
        Boom.images.append(make_boom(3, int(255)))
        Boom.images.append(make_boom(4, int(255)))
        Boom.images.append(make_boom(5, int(255)))
        Boom.images.append(make_boom(5, int(255)))
        Boom.images.append(make_boom(6, int(255)))
        Boom.images.append(make_boom(7, int(255 * 0.9)))
        Boom.images.append(make_boom(7, int(255 * 0.8)))
        Boom.images.append(make_boom(7, int(255 * 0.7)))
        Boom.images.append(make_boom(7, int(255 * 0.6)))
        Boom.images.append(make_boom(7, int(255 * 0.5)))
        Boom.images.append(make_boom(7, int(255 * 0.4)))
        Boom.images.append(make_boom(7, int(255 * 0.3)))
        Boom.images.append(make_boom(7, int(255 * 0.2)))
        Boom.image = Boom.images[0]

        # Player "station" 4x4 construct moves as a unit.
        self.player_box_color = pygame.Color('darkviolet')
        self.player_box = pygame.Rect(0, 0, 50 * 2, 50 * 2)
        self.player_box.center = settings.world_rect.center
        self.player_box_x = self.player_box.x
        self.player_box_y = self.player_box.y
        Player.player_box = self.player_box
        for i in range(4):
            self.new_player()

        self.binspace = (
            binspace.BinSpace(min_cell_size=25),
            binspace.BinHash(cell_size=25),
            binspace.GummHash(cell_size=25),
            binspacealternatives.BinSpaceEntities(cell_size=25),
        )[3]
        self.spatialhash = (
            spatialhash.SpatialHash(25),
        )[0]
        self.space = self.spatialhash   # <<< initial space container
        self.count_bin = 0
        self.count_hash = 0
        if isinstance(self.spatialhash, spatialhash.SpatialHash):
            self.spatialhash.gen_cell_ranges_cache(settings.world_rect)

        self.show_hud = True
        self.paused = False
        self.auto_fire = True
        self.escalate = False
        self.explosions = True
        self.big_baddies = True
        self.mouse_input = False
        self.mouse_pos = 0, 0
        self.movex = 0
        self.movey = 0
        self.speed = 4
        self.baddy_due = 1.0

        # Schedule player and friendly bullet spawns.
        # Note: baddy bullets are spammed in Game.update() to spread the out over each tick.
        self.player_sched = self.clock.new_schedule(self.new_player, settings.player_spawn_rate)
        self.friendly_sched = self.clock.new_schedule(self.new_friendly_bullet, settings.friendly_interval)

        self.hud = None
        self.make_hud_and_caption()

        # These are used if settings.profile is True.
        # RETURN toggles print_times during runtime.
        # See settings.very_spammy and Game.exit(); all times are dumped when the context exits.
        self.times_bin = []
        self.times_hash = []
        self.print_times = False

        # Schedule a timed quit. This is for rerunning the same load for benchmarking.
        if (settings.profile or settings.timeit) and settings.profile_duration > 0:
            def stop(*args):
                self.pop()
            self.clock.new_schedule(stop, settings.profile_duration, settings.profile_duration)

        # Schedule our own lazy GC. Python is too aggressive and causes visual stuttering.
        def gc_collect(sched):
            count = gc.get_count()
            if count[0] > 10000:
                gc.collect()
        self.gc_sched = self.clock.new_schedule(gc_collect, 10.0)

    def enter(self):
        gc.disable()

    def exit(self):
        gc.enable()
        if settings.timeit:
            nth = len(self.times_hash) or 1
            ntb = len(self.times_bin) or 1
            th = sum(self.times_hash) / nth
            tb = sum(self.times_bin) / ntb
            print('>>> Avg: Bin {:0.4f} / {:0.4f} Hash = Ratio {:0.4f}'.format(tb, th, tb / th))
            if settings.very_spammy:
                for i in range(nth):
                    h = self.times_hash[i]
                    b = self.times_bin[i]
                    print('B {:0.4f} / {:0.4f} H = {:0.4f} {}'.format(b, h, b / h, '<-' if b < h else ' ->'))

    def make_hud_and_caption(self):
        def num_sprites():
            total = len(self.baddy_bullets) + len(self.friendly_bullets) + len(self.players) + len(self.booms)
            return total, len(self.baddy_bullets), len(self.friendly_bullets), len(self.booms)

        def friendly_rate():
            bullets_per_player = 360.0 / settings.friendly_bullet_arc
            bullets_per_pulse = len(self.players) * bullets_per_player
            bullets_per_sec = bullets_per_pulse / settings.friendly_interval
            return bullets_per_sec

        def baddy_rate():
            ticks_per_sec = 1.0 / settings.time_step
            return ticks_per_sec * settings.baddy_per_tick

        def control():
            return 'Mouse' if self.mouse_input else 'Keyboard'

        def paused():
            return self.paused

        def escalate():
            return self.escalate

        def explosions():
            return self.explosions
        theme = hudthemes.HUD_THEMES['default']
        theme.update(dict(fontsize=25, owidth=0.1))
        self.hud = hudlight.HUD(**theme)
        self.hud.add('fps', 'Fps {:0.0f}', self.sched_draw.per_second(), callback=self.sched_draw.per_second)
        self.hud.add('sprites', 'Sprites {:4d} B/F/X {:4d} {:3d} {:3d}', *num_sprites(), callback=num_sprites)
        self.hud.add('friendlyrate', 'Friendly / sec {:0.1f}', friendly_rate(), callback=friendly_rate)
        self.hud.add('baddyrate', 'Baddy / sec {:0.1f}', baddy_rate(), callback=baddy_rate)
        self.hud.add('bumprate', '[+-] Spawn rate {:0.0f}'.format(1.0 / settings.time_step * settings.bump_rate))
        self.hud.add('control', '[C] Control {}', control(), callback=control)
        self.hud.add('paused', '[P] Paused {}', paused(), callback=paused)
        self.hud.add('escalate', '[E] Escalate {}', escalate(), callback=escalate)
        self.hud.add('explosions', '[X] Explosions {}', explosions(), callback=explosions)
        self.clock.new_schedule(self.hud.update, 0.5)

        def caption(*args):
            fmt = 'Spacebin{prof} | Space: Fire! | H> HUD {showhud} | F> Auto {autofire} | Tab> {space} |' + \
                  ' Fps {fps:0.0f} | Sprites {nsprites} | Baddy / sec {baddyrate:0.1f}'
            pygame.display.set_caption(fmt.format(
                prof=' (Profiling)' if settings.profile else '',
                showhud='On' if self.show_hud else 'Off',
                autofire='On' if self.auto_fire else 'Off',
                space=self.space.__class__.__name__,
                fps=self.sched_draw.per_second(), nsprites=num_sprites(), baddyrate=baddy_rate()))
        self.clock.new_schedule(caption, 1.0)

    def new_player(self, sched=None):
        if len(self.players) > 3:
            return
        number = 0
        for i, s in sorted([(s.number, s) for s in self.players]):
            if i != number:
                break
            number += 1
        rect = Player.image.get_rect()
        corner = Player.corners[number]
        setattr(rect, corner, getattr(self.player_box, corner))
        sprite = Player(rect)
        sprite.number = number
        self.players.append(sprite)

    def new_baddy_bullet(self, sched=None):
        step = settings.baddy_per_tick
        self.baddy_due += step
        if self.baddy_due < 24.0:
            return

        baddy_due = self.baddy_due
        self.baddy_due %= step

        bullets_append = self.baddy_bullets.append
        dt = self.clock.period
        size = 5
        base_speed = 20 * dt
        half_speed = base_speed / 2.0
        for i in range(int(baddy_due)):
            side = random.choice(('down', 'left', 'up', 'right'))
            if side == 'down':
                x = random.randrange(settings.world_rect.w)
                y = size
                dirx = random.random() * base_speed - half_speed
                diry = random.random() * half_speed + half_speed
            elif side == 'left':
                x = settings.world_rect.w - size
                y = random.randrange(settings.world_rect.h)
                dirx = random.random() * -half_speed + -base_speed
                diry = random.random() * base_speed - half_speed
            elif side == 'up':
                x = random.randrange(settings.world_rect.w)
                y = settings.world_rect.bottom - size
                dirx = random.random() * base_speed - half_speed
                diry = random.random() * -half_speed + -half_speed
            else:  # side == 'right'
                x = size
                y = random.randrange(settings.world_rect.h)
                dirx = random.random() * half_speed + half_speed
                diry = random.random() * base_speed - half_speed
            if not self.big_baddies or random.random() > settings.big_baddy_ratio:
                rect = BaddyBullet.image.get_rect(center=(x, y))
                sprite = BaddyBullet(rect)
            else:
                img = random.choice(BigBaddyBullet.images)
                rect = img.get_rect(center=(x, y)).clamp(settings.screen_rect)
                sprite = BigBaddyBullet(rect)
                sprite.image = img
                sprite.reset_image = img
            sprite.speed = 1
            sprite.dirx = dirx
            sprite.diry = diry
            bullets_append(sprite)

    def new_friendly_bullet(self, sched=None):
        # ignore scheduled auto fire if it is turned off
        if sched and not self.auto_fire:
            return

        dt = self.clock.period
        speed = 75.0
        for player in self.players:
            cx, cy = player.rect.center
            r = player.rect.w // 2
            for angle in range(0, 360, settings.friendly_bullet_arc):
                    x, y = point_on_circumference((cx, cy), r, angle)
                    rect = FriendlyBullet.image.get_rect(center=(x, y))
                    sprite = FriendlyBullet(rect)
                    dirx, diry = point_on_circumference((0, 0), speed * dt, angle + 45.0 / 2.0)
                    sprite.dirx = dirx
                    sprite.diry = diry
                    self.friendly_bullets.append(sprite)

    def new_boom(self, pos, vec=None):
        rect = Boom.image.get_rect(center=pos)
        sprite = Boom(rect)
        if vec:
            sprite.dirx = vec[0]
            sprite.diry = vec[1]
        self.booms.append(sprite)

    def update(self, dt):
        if self.paused:
            self.update_paused()
        else:
            self.get_events()
            self.update_sprites(dt)
            self.update_collisions(dt)

    def update_paused(self):
        e = pygame.event.wait()
        if e.type == pygame.KEYDOWN and e.key == pygame.K_p:
            self.paused = False
            self.clock.paused = False

    def update_sprites(self, dt):
        if self.mouse_input:
            self.player_box.center = self.mouse_pos
        elif self.movex or self.movey:
            self.player_box_x += self.movex
            self.player_box_y += self.movey
            self.player_box.topleft = self.player_box_x, self.player_box_y

        self.new_baddy_bullet()

        baddy_dead = []
        friendly_dead = []
        player_dead = []
        boom_dead = []
        cold_boom_dead = []
        sprite_groups = self.baddy_bullets, self.friendly_bullets, self.players, self.booms, self.cold_booms
        dead_groups = baddy_dead, friendly_dead, player_dead, boom_dead, cold_boom_dead
        for dead, group in zip(dead_groups, sprite_groups):
            for s in group:
                s.update(dt)
                if s.lives <= 0:
                    dead.append(s)

        for dead, group in zip(dead_groups, sprite_groups):
            for s in dead:
                group.remove(s)
        boom_spread_end = settings.boom_spread_end
        cold_booms = [s for s in self.booms if s.elapsed > boom_spread_end]
        for s in cold_booms:
            self.booms.remove(s)
        self.cold_booms.extend(cold_booms)

    def timeit(self, stmt, name):
        time_spent = timeit.timeit(stmt, globals=locals(), number=1)
        getattr(self, name).append(time_spent)
        return time_spent

    def update_collisions(self, dt):
        if settings.timeit:
            if self.space is self.spatialhash:
                self.timeit('self.collide_bin()', 'times_bin')
                self.timeit('self.collide_hash()', 'times_hash')
                assert self.count_hash == self.count_bin, 'B {} != {} H'.format(self.count_bin, self.count_hash)
            else:
                self.timeit('self.collide_hash()', 'times_hash')
                self.timeit('self.collide_bin()', 'times_bin')
                assert self.count_hash == self.count_bin, 'B {} != {} H'.format(self.count_bin, self.count_hash)
            if self.print_times and self.times_hash:
                h = self.times_hash[-1]
                b = self.times_bin[-1]
                print('Bin {:0.5f} {:0.5f} {:0.5f} Hash'.format(b, b / h, h))
        elif self.space is self.binspace:
            self.collide_bin()
        else:
            self.collide_hash()

        dead_baddy = self.dead_baddy
        dead_baddy_boom = self.dead_baddy_boom
        dead_friendly = self.dead_friendly
        dead_player = self.dead_player

        if self.escalate and (dead_baddy or dead_baddy_boom):
            settings.baddy_per_tick += settings.time_step / 100

        baddy_bullets = self.baddy_bullets
        new_boom = self.new_boom
        bb_remove = self.baddy_bullets.remove
        fb_remove = self.friendly_bullets.remove
        p_remove = self.players.remove
        if self.explosions:
            for s in chain(dead_baddy, dead_baddy_boom):
                if isinstance(s, BigBaddyBullet):
                    r = s.rect
                    center = r.center
                    vecs = [Vec2d(point_on_circumference(center, 25, angle_of(center, p))) - center for p in
                            (r.topleft, r.midtop, r.topright, r.midright,
                             r.bottomright, r.midbottom, r.bottomleft, r.midleft)]
                    for v in vecs:
                        new_boom(center, v)
                else:
                    center = s.rect.center
                    for i in range(random.randrange(2, 5)):
                        angle = random.randrange(360)
                        v = Vec2d(point_on_circumference(center, random.randrange(5, 15), angle)) - center
                        new_boom(center, v)
        [bb_remove(s) for s in dead_baddy]
        [bb_remove(s) for s in dead_baddy_boom if s in baddy_bullets]
        [fb_remove(s) for s in dead_friendly]
        [p_remove(s) for s in dead_player]
        self.dead_baddy_boom.clear()

    def collide_bin(self):
        def collate_b(coll, deada, deadb):
            # Code for bins that don't do colliderect:
            # deada_add = deada.add
            # for a in coll:
            #     collide = a.rect.colliderect
            #     colls = set([b for b in coll[a] if collide(b.rect)])
            #     deadb |= colls
            #     if colls:
            #         deada_add(a)
            deada |= set(coll)
            deadb |= set(chain.from_iterable(coll.values()))

        dead_baddy = self.dead_baddy
        dead_baddy_boom = self.dead_baddy_boom
        dead_friendly = self.dead_friendly
        dead_player = self.dead_player
        [s.clear() for s in (dead_baddy, dead_friendly, dead_player)]

        self.count_bin = 0
        space = self.binspace
        space_collides = space.collides
        space.clear()
        boom_spread_start = settings.boom_spread_start
        boom_spread_end = settings.boom_spread_end
        if isinstance(space, binspace.GummHash):
            space.addlist(self.baddy_bullets)
            collate_b(space_collides(self.friendly_bullets), dead_friendly, dead_baddy)
            collate_b(space_collides(self.players), dead_player, dead_baddy)
            collate_b(space_collides([s for s in self.booms if boom_spread_start <= s.elapsed <= boom_spread_end]),
                      set(), dead_baddy_boom)
        else:
            space.adds_with_radius(self.baddy_bullets)
            collate_b(space_collides(self.friendly_bullets), dead_friendly, dead_baddy)
            collate_b(space_collides(self.players), dead_player, dead_baddy)
            collate_b(space_collides([s for s in self.booms if boom_spread_start <= s.elapsed <= boom_spread_end]),
                      set(), dead_baddy_boom)
        self.count_bin += sum([len(d) for d in (dead_baddy, dead_baddy_boom, dead_friendly, dead_player)])

    def collide_hash(self):
        def collate_h(coll, deada, deadb):
            deada |= set(coll)
            deadb |= set(chain.from_iterable(coll.values()))

        dead_baddy = self.dead_baddy
        dead_baddy_boom = self.dead_baddy_boom
        dead_friendly = self.dead_friendly
        dead_player = self.dead_player
        [s.clear() for s in (dead_baddy, dead_friendly, dead_player)]

        self.count_hash = 0
        space = self.spatialhash
        space.clear()
        space.addlist(self.baddy_bullets)
        space_collides = space.collidedict
        boom_spread_start = settings.boom_spread_start
        boom_spread_end = settings.boom_spread_end
        collate_h(space_collides(self.friendly_bullets, do_fine=False), dead_friendly, dead_baddy)
        collate_h(space_collides(self.players, do_fine=False), dead_player, dead_baddy)
        collate_h(space_collides([s for s in self.booms if boom_spread_start <= s.elapsed <= boom_spread_end],
                                 do_fine=False), set(), dead_baddy_boom)
        self.count_hash += sum([len(d) for d in (dead_baddy, dead_baddy_boom, dead_friendly, dead_player)])

    def draw(self, sched):
        screen = settings.screen
        screen.fill(self.bgcolor)
        pygame.draw.rect(screen, self.player_box_color, self.player_box, 1)
        self.draw_sprites(screen)
        if self.show_hud:
            self.hud.draw(screen)
        pygame.display.flip()

    def draw_sprites(self, screen):
        blit = screen.blit
        for g in (self.friendly_bullets, self.players, self.baddy_bullets, self.cold_booms, self.booms):
            for s in g:
                blit(s.image, s.rect)

    def on_mouse_motion(self, pos, rel, buttons):
        self.mouse_pos = pos

    def on_key_down(self, unicode, key, mod):
        if key == pygame.K_ESCAPE:
            self.pop()
        elif key == pygame.K_SPACE:
            self.new_friendly_bullet()
        elif key == pygame.K_RETURN:
            self.print_times = not self.print_times
        elif key == pygame.K_TAB:
            # switch between binspace and spatialhash
            self.space = self.binspace if self.space is self.spatialhash else self.spatialhash
        elif key in (pygame.K_PLUS, pygame.K_EQUALS, pygame.K_KP_PLUS):
            # bump rate up
            settings.baddy_per_tick += settings.bump_rate
        elif key in (pygame.K_MINUS, pygame.K_KP_MINUS):
            # bump rate down
            settings.baddy_per_tick -= settings.bump_rate
            if settings.baddy_per_tick <= 0.5:
                settings.baddy_per_tick = 0.5
        elif key == pygame.K_b:
            # toggle big baddies
            self.big_baddies = not self.big_baddies
        elif key == pygame.K_c:
            # switch between mouse and keyboard control
            self.mouse_input = not self.mouse_input
            if self.mouse_input:
                self.player_box.center = self.mouse_pos
            else:
                self.player_box_x, self.player_box_y = self.player_box.topleft
        elif key == pygame.K_e:
            # toggle escalate
            self.escalate = not self.escalate
        elif key == pygame.K_f:
            # toggle auto fire
            self.auto_fire = not self.auto_fire
        elif key == pygame.K_h:
            self.show_hud = not self.show_hud
        elif key == pygame.K_p:
            # toggle pause
            self.paused = True
            self.clock.paused = True
        elif key == pygame.K_x:
            # toggle explosions
            self.explosions = not self.explosions
        elif key in (pygame.K_d, pygame.K_RIGHT):
            self.movex += 1
        elif key in (pygame.K_a, pygame.K_LEFT):
            self.movex += -1
        elif key in (pygame.K_s, pygame.K_DOWN):
            self.movey += 1
        elif key in (pygame.K_w, pygame.K_UP):
            self.movey += -1

    def on_key_up(self, key, mod):
        if key in (pygame.K_d, pygame.K_RIGHT):
            self.movex += -1
        elif key in (pygame.K_a, pygame.K_LEFT):
            self.movex += 1
        elif key in (pygame.K_s, pygame.K_DOWN):
            self.movey += -1
        elif key in (pygame.K_w, pygame.K_UP):
            self.movey += 1

    def on_quit(self):
        self.pop()
