"""settings.py - startup and runtime settings file

Usage
=====
Settings in this file should always be accessed like so:
    
    import settings
    if settings.whatever: ...
    settings.whatever = value

This insures that runtime changes will be noticed by other modules that import
the settings module. Of course, one can always make local copies of the values
if static behavior is desired.
"""

import pygame

import imagecache
import pygametext


class DictToSettings(object):
    """convert an arbitrarily nested dict to an object of nested attributes with the same hierarchy

    Caveat: does not work if the final value is a dict. :) In such a case do:

        thing = Setting(**dict(key1=dict(key2=None)))
        thing.key1.key2 = dict(...)
    """
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if isinstance(value, dict):
                value = DictToSettings(**value)
            setattr(self, key, value)


class Setting(object):
    """easily build an object with nested attributes
    thing = Setting('key1.key2.key3', 123)
    thing = Setting()
    thing.key1.key2.key3 = 123
    print(thing.key1.key2.key3)
    """
    def __init__(self, attr=None, value=None):
        if isinstance(attr, str):
            setattr(self, attr, value)

    def __setattr__(self, attr, value):
        attrs = attr.split('.')
        if len(attrs) == 1:
            self.__dict__[attrs[0]] = value
        elif len(attrs) > 1:
            self.__dict__[attrs[0]] = Setting('.'.join(attrs[1:]), value)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            self.__dict__[attr] = Setting()
            return self.__dict__[attr]


# CPU
time_step = 1.0 / 30.0              # recommended: 30, 60, ...
draw_step = 1.0 / 150.0             # recommended: 50, 150, 200, 250, ...

# Display
resolution = 1024, 768              # screen size in pixels: width,height
display_flags = pygame.DOUBLEBUF    # pygame.FULLSCREEN
screen = None                       # set me: when screen is opened or changed
screen_rect = None                  # set me: when screen is opened or changed

# Game details
world_rect = pygame.Rect((0, 0), resolution)
tile_size = 16, 16

# Game behavior
baddy_per_tick = time_step / (1.0 / 120.0)  # <<< n per tick, equivalent to 30 per sec
bump_rate = time_step / (1.0 / 15.0)        # <<< +/- bump spawn rate up or down
big_baddy_ratio = 0.05                      # <<< random() > this == big baddy bullet

friendly_bullet_arc = 12    # <<< size of the arc between guns in degrees, e.g. 8, 10, 12, 15, 20
friendly_interval = 4.0     # <<< fire bullet spray every N seconds
player_spawn_rate = 4.0     # <<< check to respawn player evern N seconds

# Graphics
image_dir = 'data/image'
pygametext.FONT_NAME_TEMPLATE = 'data/font/%s'
boom_spread_start = (0.33, 0.1)[1]    # elapsed time of a boom to start blowing up neighbors
boom_spread_end = (0.44, 0.8)[1]      # elapsed time of a boom to stop blowing up neighbors

# cached surfaces (required by spritefx)
# Note: pygametext does its own caching. No need to cache surfaces from pygametext.getsurf() unless you transform them.
# Note: these settings are not tunable on the fly. If you want to change the image cache settings during runtime, you'll
# have to call the image_cache instance's methods.
image_cache_expire = 1 * 60          # expire unused textures after N seconds; recommend 1 min (1 * 60)
image_cache_memory = 5 * 2 ** 20     # force expiration if memory is exceeded; recommend 5 MB (5 * 2**20)
image_cache = imagecache.ImageCache(image_cache_expire, image_cache_memory)
image_cache.enable_age_cap(True)

# font_themes is used with the functions in module pygametext
# Example call to render some text using a theme:
#   image = pygametext.getsurf('some text', **settings.font_themes['intro'])
# font_themes can be accessed by nested dict keys:
#   font_theme['intro']['fontname']
font_themes = dict(
    intro=dict(
        fontname='Boogaloo.ttf',
        fontsize=52,
        color='red2',
        gcolor='deepskyblue',
        ocolor='yellow',
        owidth=0.5,
        scolor=None,
        shadow=None,
    ),
    mainmenu=dict(
        fontname='Bubblegum_Sans.ttf',
        fontsize=80,
        color='red2',
        gcolor='orange1',
        ocolor='yellow',
        owidth=0.5,
        scolor=None,
        shadow=None,
    ),
    game=dict(
        fontname='Vera.ttf',
        fontsize=38,
        color='lightblue',
        gcolor='green4',
        ocolor=None,
        owidth=0.5,
        scolor='grey20',
        shadow=(-1, 1),
    ),
    floatingtext=dict(
        fontname='Vera.ttf',
        fontsize=18,
        color='red',
        gcolor='lightblue4',
        ocolor='yellow',
        owidth=0.5,
        scolor=None,
        shadow=None,
    ),
)
# font_info can be accessed by nested attributes:
#   font_info.intro.fontname
font_info = DictToSettings(**font_themes)


# Sound
sound_dir = 'data/sound'
music_dir = 'data/music'

# Debug
profile = False             # run cProfile if True
timeit = False              # collect timings
very_spammy = False         # include very_spammy output
profile_duration = 30.0     # time in seconds until auto-quit, only relevant if profile is True

# Custom
import os
os.environ['SDL_VIDEO_CENTERED'] = '1'
bypass_menu = None

# Gummbum's development preferences... remove or add a case for your own.
_who = 'nobody'
_whos = 'gumm dr0id nobody'.split()
for _who in _whos:
    if os.access('iam' + _who, os.F_OK):
        break
try:
    if _who == 'gumm':
        print('## CUSTOM SETTINGS FOR gumm (see settings.py)')
        # See main.py for the effect of bypass_menu:
        bypass_menu = (None, 'game', 'credits')[1]
        profile, timeit, very_spammy = False, False, False
        profile_duration = (20.0, 30.0, 60.0, 180.0)[0]
        if profile or timeit:
            import random
            random.seed(12345)
        if very_spammy:
            print('very_spammy output sent to gumm.txt')
            import pipestdio
            pipestdio.stdout.to_logfile('gumm.txt', 'w')
            del pipestdio

    elif _who == 'dr0id':
        print('## CUSTOM SETTINGS FOR dr0id (see settings.py)')
        pass
except Exception as e:
    print('## exception in custom settings: ignoring\n{}'.format(e))

del os, pygame, _who, _whos
