"""convenient classes to pipe or restore stdout and stderr

Contents of this module:
    PipeStdout  redirect sys.stdout to console, devnull, or a file
    PipeStderr  redirect sys.stderr to console, devnull, or a file
    TeeStdout   redirect sys.stdout to console and a file
    TeeStderr   redirect sys.stderr to console and a file
    restore()   restore stdout and stderr to their original streams
    flush()     flush sys.stdout and sys.stderr
    stdout      module-level instance of PipeStdout
    stderr      module-level instance of PipeStderr

Example 1, no stdout, and stderr to logfile:
stdout.to_devnulldev()
stderr.to_logfile('mylog.txt')

Example 2, stdout and stderr same logfile:
f = open('mylog.txt', 'wb')
stdout.to_fh(f)
stderr.to_fh(f)
-or-
stdout.to_logfile('mylog.txt')
stderr.to_fh(stdout.get_current_fh())
-or-
stdout.to_logfile('mylog.txt')
stderr.to_logfile('mylog.txt', mode='w+b')

Example 3, dupe output to console and logfile:
tee = TeeStdout('mylog.txt', 'wb')
stdout.to_fh(tee)
stderr.to_fh(tee)
"""

import os
import sys


_STDOUT = sys.stdout
_STDERR = sys.stderr


class _Pipe(object):
    _sysout = None
    _attr = None

    def restore(self):
        setattr(sys, self._attr, self._sysout)
        return self

    def get_sys_fh(self):
        return self._sysout

    def get_current_fh(self):
        return getattr(sys, self._attr)

    def flush(self):
        self._sysout.flush()

    def to_devnull(self):
        # if sys.platform == 'win32':
        #     f = open('NUL:', 'wb')
        # else:
        #     f = open('/dev/null', 'wb')
        f = open(os.devnull, 'w')
        setattr(sys, self._attr, f)
        return self

    def to_logfile(self, name, mode='wb'):
        f = open(name, mode)
        setattr(sys, self._attr, f)
        return self

    def to_fh(self, f):
        setattr(sys, self._attr, f)
        return self


class PipeStdout(_Pipe):
    _sysout = _STDOUT
    _attr = 'stdout'


class PipeStderr(_Pipe):
    _sysout = _STDERR
    _attr = 'stderr'


class _Tee(object):
    _sysout = None
    _attr = None

    def __init__(self, name_or_fh, mode='wb'):
        if isinstance(name_or_fh, str):
            self.file = open(name_or_fh, mode)
        else:
            self.file = name_or_fh
        self.console = self._sysout
        setattr(sys, self._attr, self)

    def __del__(self):
        self.close()

    def write(self, data):
        self.file.write(data)
        self.console.write(data)

    def flush(self):
        self.file.flush()
        self.console.flush()

    def close(self):
        setattr(sys, self._attr, self._sysout)
        self.file.close()


class TeeStdout(_Tee):
    _sysout = _STDOUT
    _attr = 'stdout'

    def __init__(self, name, mode):
        _Tee.__init__(self, name, mode)


class TeeStderr(_Tee):
    _sysout = _STDERR
    _attr = 'stderr'

    def __init__(self, name, mode):
        _Tee.__init__(self, name, mode)


def restore():
    flush()
    stdout.restore()
    stderr.restore()


def flush():
    stdout.flush()
    stderr.flush()


stdout = PipeStdout()
stderr = PipeStderr()


if __name__ == '__main__':
    # TODO: test cases
    def write(s):
        sys.stdout.write(s)
        flush()

    def write2(s):
        sys.stdout.write('{}: to stdout\n'.format(s))
        sys.stderr.write('{}: to stderr\n'.format(s))
        flush()

    def cat(name):
        with open(name, mode='rb') as f:
            for s in f:
                write('cat: ' + s)
        flush()

    write('\ntest 0a: out=devnull err=devnull\n')
    stdout.to_devnull()
    stderr.to_devnull()
    write2('test 0a')
    restore()

    write('\ntest 0b: out=console err=console\n')
    write2('test 0b')
    restore()

    write('\ntest 1: out=devnull err=logfile\n')
    stdout.to_devnull()
    stderr.to_logfile('myfile.txt')
    write2('test 1')
    stderr.get_current_fh().close()
    restore()
    cat('myfile.txt')

    write('\ntest 2a: out=logfile err=logfile\n')
    f = open('myfile.txt', 'wb')
    stdout.to_fh(f)
    stderr.to_fh(f)
    write2('test 2a')
    f.close()
    restore()
    cat('myfile.txt')

    write('\ntest 2b: out=logfile err=logfile\n')
    stdout.to_logfile('mylog.txt')
    stderr.to_fh(stdout.get_current_fh())
    write2('test 2b')
    stdout.get_current_fh().close()
    restore()
    cat('myfile.txt')

    write('\ntest 2c: out=logfile err=logfile\n')
    stdout.to_logfile('mylog.txt')
    stderr.to_logfile('mylog.txt', mode='w+b')
    write2('test 2c')
    stdout.get_current_fh().close()
    restore()
    cat('myfile.txt')

    write('\ntest 3: out=tee err=tee\n')
    tee = TeeStdout('myfile.txt', 'wb')
    stdout.to_fh(tee)
    stderr.to_fh(tee)
    write2('test 3')
    stdout.get_current_fh().close()
    restore()
    cat('myfile.txt')

    write('\n')
    write2('done')
    write('\n')
