import pygame

import pygametext
import settings
import tweening
from vec2d import Vec2d


tweener = tweening.Tweener()
texts = set()


def update_tweens(dt, update_image_cache=True):
    """update the tweens and the image cache
    
    If you aren't using any of the tween effects, you don't need to call this.
    
    This function updates settings.image_cache. If you want to update the image cache elsewhere, then pass in
    update_image_cache=False.
    
    :param dt: time slice from the clock
    :param update_image_cache: update settings.image_cache or not
    :return: None
    """
    tweener.update(dt)
    for t in tweener.get_all_tweens():
        if t.o._dirty:
            t.o._refresh_image()
    if update_image_cache:
        settings.image_cache.update()


def clear(do_tweens=True, do_texts=True):
    if do_tweens:
        tweener.clear()
    if do_texts:
        texts.clear()


class SpriteFX(object):
    """a sprite class that provides a number of managed tween effects and interpolated rendering
    """
    layer = 0

    def __init__(self, image, pos, angle=0.0, scale=1.0, alpha=255, colorkey=None, flip=None, anchor='topleft', layer=None, interp=True):
        self.image = image
        self.rect = image.get_rect(**{anchor: pos})
        self.interp = interp
        self._source_image = image
        self._pos = Vec2d(pos)
        self._old_pos = Vec2d(pos)
        self._move_to = Vec2d(pos)
        self._angle = angle
        self._scale = scale
        self._alpha = alpha
        self._colorkey = colorkey
        self._anchor = anchor
        self._flip = flip
        if layer is not None:
            self.layer = layer

        self.tweens = {}

        self._dirty = True
        self._dirty_transparent = True
        self._dirty_pos = True

    def _get_angle(self):
        return self._angle
    def _set_angle(self, value):
        if value != self._angle:
            self._angle = value
            self._dirty = True
    angle = property(_get_angle, _set_angle)
        
    def _get_scale(self):
        return self._scale
    def _set_scale(self, value):
        if value != self._scale:
            self._scale = value
            self._dirty = True
    scale = property(_get_scale, _set_scale)

    def _get_alpha(self):
        return self._alpha
    def _set_alpha(self, value):
        self._alpha = value
        self.image.set_alpha(value)
        self._dirty_transparent = True
    alpha = property(_get_alpha, _set_alpha)

    def _get_colorkey(self):
        return self._colorkey
    def _set_colorkey(self, value):
        self._angle = value
        self.image.set_colorkey(value)
        self._dirty_transparent = True
    colorkey = property(_get_colorkey, _set_colorkey)

    def _get_anchor(self):
        return self._anchor
    def _set_anchor(self, value):
        old_pos = getattr(self.rect, self._anchor)
        new_pos = getattr(self.rect, value)
        diff = Vec2d(new_pos) - old_pos
        self._anchor = value
        self._pos += diff
    anchor = property(_get_anchor, _set_anchor)

    def _get_flip(self):
        return self._flip
    def _set_flip(self, value):
        if value != self._flip:
            self._flip = value
            self._dirty = True
    flip = property(_get_flip, _set_flip)

    def _get_pos(self):
        return self._pos
    def _set_pos(self, val):
        self._pos[:] = val
        setattr(self.rect, self._anchor, val)
        self._dirty_pos = True
    pos = property(_get_pos, _set_pos)

    def _get_x(self):
        return self.pos.x
    def _set_x(self, value):
        self._update_pos((value, self._pos.y), self.interp)
    x = property(_get_x, _set_x)

    def _get_y(self):
        return self.pos.y
    def _set_y(self, value):
        self._update_pos((self._pos.x, value), self.interp)
    y = property(_get_y, _set_y)

    def _get_move_to(self):
        return self._move_to
    def _set_move_to(self, val):
        self._move_to[:] = val
    move_to = property(_get_move_to, _set_move_to)

    def _get_old_pos(self):
        return self._old_pos
    def _set_old_pos(self, val):
        self._old_pos[:] = val
    old_pos = property(_get_old_pos, _set_old_pos)

    def _update_pos(self, val, interp=True):
        if not self._dirty_pos:
            self.old_pos = self.pos if interp else val
        self.pos = self.move_to = val
        setattr(self.rect, self.anchor, val)

    def hit(self, *args):
        """override"""
        pass

    def kill(self, *args):
        """override"""
        pass

    def update(self, dt):
        """override"""
        pass

    def draw(self, surface, interp=1.0):
        if self._dirty or self._dirty_transparent:
            self._refresh_image()
        rect = self.rect
        if 0.0 <= interp < 1.0:
            # interpolate position
            fm = self.old_pos
            to = self.pos
            if fm != to:
                dx = (to.x - fm.x) * interp
                dy = (to.y - fm.y) * interp
                rect = rect.move(dx, dy)
        surface.blit(self.image, rect)

    def _refresh_image(self):
        bucket_key = id(self._source_image)
        transforms = self._angle, self._scale, self._flip, self._alpha, self._colorkey
        cached_image = settings.image_cache.get(bucket_key, transforms)
        if cached_image:
            self.image = cached_image
        else:
            if self.angle != 0.0 or self.scale != 1.0:
                self.image = pygame.transform.rotozoom(self._source_image, self.angle, self.scale)
            else:
                self.image = self._source_image.copy()
            if self.flip:
                self.image = pygame.transform.flip(self.image, *self.flip)
            if self.image.get_flags() & pygame.SRCALPHA:
                self.image.fill((255, 255, 255, self._alpha), None, pygame.BLEND_RGBA_MULT)
            else:
                self.image.set_alpha(self._alpha)
                self.image.set_colorkey(self._colorkey)
            settings.image_cache.add(bucket_key, transforms, self.image)
        diff = self.pos - getattr(self.rect, self.anchor)
        center = self.rect.center
        self.rect = self.image.get_rect(center=center)
        new_pos = getattr(self.rect, self.anchor)
        self.pos = diff + new_pos
        self._dirty = False
        self._dirty_transparent = False

    def fade_effect(self, start_alpha=255, end_alpha=128, duration=1.0, tween_function=tweening.ease_linear, cb_end=None):
        self.misc_effect('alpha', start_alpha, end_alpha, duration, tween_function, cb_end)
        return self

    def spin_effect(self, start_angle=0.0, end_angle=360.0, duration=1.0, tween_function=tweening.ease_linear, cb_end=None):
        self.misc_effect('angle', start_angle, end_angle, duration, tween_function, cb_end)
        return self

    def scale_effect(self, start_scale=1.0, end_scale=2.0, duration=1.0, tween_function=tweening.ease_linear, cb_end=None):
        self.misc_effect('scale', start_scale, end_scale, duration, tween_function, cb_end)
        return self

    def driftx_effect(self, start_x=0.0, end_x=0.0, duration=1.0, tween_function=tweening.ease_linear, cb_end=None):
        self.misc_effect('x', start_x, end_x, duration, tween_function, cb_end)
        return self

    def drifty_effect(self, start_y=0.0, end_y=0.0, duration=1.0, tween_function=tweening.ease_linear, cb_end=None):
        self.misc_effect('y', start_y, end_y, duration, tween_function, cb_end)
        return self

    def misc_effect(self, attr_name, start_val, end_val, duration, tween_function=tweening.ease_linear, cb_end=None):
        """

        :param attr_name: str; the attribute created on this object to store the computed value
        :param start_val: int or float; starting value for the tween
        :param end_val: int or float; ending value for the tween
        :param duration: int or float; duration of the tween in seconds
        :param tween_function: see tweening module for available functions
        :param cb_end: list; callback args; see prite._cb_end() for handling the callback args
        :return:
        """
        if cb_end is None:
            cb_end = self.tween_callback_end
        self.tweens[attr_name] = tweener.create_tween(
            self, attr_name, start_val, end_val - start_val, duration,
            tween_function, cb_end=cb_end, cb_args=(self, attr_name))
        return self

    def end_fade(self):
        """preempt and remove a tween"""
        self.end_misc('alpha')

    def end_spin(self):
        """preempt and remove a tween"""
        self.end_misc('angle')

    def end_scale(self):
        """preempt and remove a tween"""
        self.end_misc('scale')

    def end_driftx(self):
        """preempt and remove a tween"""
        self.end_misc('driftx')
        self.driftx = 0.0

    def end_drifty(self):
        """preempt and remove a tween"""
        self.end_misc('drifty')
        self.drifty = 0.0

    def end_misc(self, attr_name):
        """preempt and remove a tween"""
        if attr_name in self.tweens:
            tweener.remove_tween(self.tweens[attr_name])
            del self.tweens[attr_name]

    def tween_callback_end(self, sprite, attr_name, tween):
        """called when a tween ends, cleans up the tween
        If you supply a different callback when starting an effect, make sure to call this one or duplicate its
        function.
        
        :param sprite: sprite as returned by the tweener
        :param attr_name: attr_name  as returned by the tweener
        :param tween: tween as returned by the tweener
        :return: None
        """
        # args: see Sprite.misc_effect() and tweening.Tweener.update()
        if attr_name in self.tweens:
            del self.tweens[attr_name]


class TextSprite(SpriteFX):
    """a sprite class that provides text rendering, a number of managed tween effects and interpolated rendering 
    """
    layer = 0

    def __init__(self, text, pos, expire=None, theme_name='floating_text', angle=0.0, scale=1.0, alpha=255, colorkey=None, flip=None, anchor='topleft', layer=None, interp=True):
        """create a text sprite
        
        expire is the duration that the text is kept alive. When sprite.alive is False, the sprite is auto-removed from
        the internal texts list. The alive attribute should also be checked if the sprite is being managed externally.
        If expire is None then the text sprite is kept forever, requiring manual removal from spritefx.texts or
        spritefx.clear(do_texts=True).
        
        theme_name is the key in the settings.font_theme dict which provides the arguments for pygametext.getsurf().
        See settings.font_themes. Example theme:
            settings.font_themes = {
                'vera': {
                    'fontname': 'Vera.ttf',
                    'fontsize': 18,
                    'color': 'gold',
                    'gcolor': 'orange3',
                    'ocolor': 'black',
                    'owidth': 0.5,
                }
            }
        
        Except for text, expire, and theme_name this class accepts all the same arguments as SpriteFX. Please refer to
        SpriteFX for a further explanation of the remaining params.
        """
        self.alive = True
        self.text = text

        if layer is not None:
            self.layer = layer

        if alpha != 255:
            surf = pygametext.getsurf(text, alpha=1.0 / alpha if alpha > 0 else 0,
                                      **settings.font_themes[theme_name]).convert()
            surf.set_alpha(alpha)
        else:
            surf = pygametext.getsurf(text, **settings.font_themes[theme_name])
        super(TextSprite, self).__init__(surf, pos, angle, scale, alpha, colorkey, flip, anchor, layer, interp)

        if expire is not None:
            self.misc_effect('_expire', 1, 1, expire, cb_end=self.__expired)

        texts.add(self)

    def __expired(self, *args):
        self.tween_callback_end(*args)
        self.remove()

    def remove(self):
        self.alive = False
        if self in texts:
            texts.remove(self)
